#include <gwin.h>

void Gwin::drawline (int h1, int v1, int h2, int v2)
{
    register int
    	xmin = border_present,
    	xmax = width - border_present,
    	ymin = border_present,
    	ymax = height - border_present;
    	
    // let's make sure that the line increases in the horizontal 
    // direction; i.e., from h1 to h2 
    if (h1 > h2)
    {
    	register int
    	    tmp;
    	
    	tmp = h1;
    	h1 = h2;
    	h2 = tmp;
    	tmp = v1;
    	v1 = v2;
    	v2 = tmp;
    }
    	
    // if h1 and h2 are outside the window in the same sector.. or
    // if v1 and v2 are outside the window in the same sector, nogo
    if ( (h1 > xmax)		  ||
    	 (h2 < xmin)		  ||
    	 (v1 < ymin && v2 < ymin) ||
    	 (v1 > ymax && v2 > ymax)
       )
	return;
	
    // if all coordinates are within the window, simply plot
    if ( (h1 >= xmin)		    &&
         (h2 <= xmax)		    &&
         (v1 >= ymin && v1 <= ymax) &&
         (v2 >= ymin && v2 <= ymax)
       ) 
    {    
    	Graph::setcolor (foreground_color);
    	Graph::drawline (x1 + h1, y1 + v1, x1 + h2, y1 + v2);
    	return;
    }
    
    // next, let's see if this a vertical line.. if so, we handle
    // it as a special case
   int
   	hone,					// coordinates are now
   	vone,					// remapped to 
   	htwo,					// (hone, vone) and
   	vtwo;					// (htwo, vtwo)
    
    if (h1 == h2)
    {
    	if (v1 < ymin)
    	    vone = ymin;
    	else if (v1 > ymax)
    	    vone = ymax;
    	else
    	    vone = v1;
    	    
    	if (v2 < ymin)
    	    vtwo = ymin;
    	else if (v2 > ymax)
    	    vtwo = ymax;
	else
	    vtwo = v2;
    	    
    	Graph::setcolor (foreground_color);
    	Graph::drawline (x1 + h1, y1 + vone, x1 + h1, y1 + vtwo);
    	
    	return;
    }
    
    // we'll have to clip, so let's determine the line formula
    // given by v = ah + b; where a is the slope of the line and
    // b is the intersection with the v-axis at h == 0
    double
    	a,
    	b;
    	
    a = ((double) v2 - (double) v1) / 		// slope is dy/dx
        ((double) h2 - (double) h1);
	    
    b = v1 - a * h1;

    // Next, the left and right starting points are computed. There's some
    // redundancy of code here... but duplicating the code is hopefully
    // advantageous to speed..
   
    // determine left starting point
    if ( (h1 >= xmin && h1 <= xmax) &&		// first estimate:
    	 (v1 >= ymin && v1 <= ymax)		// maybe we can use (h1,v1)
       )					// as left starting point?
    {
    	hone = h1;
    	vone = v1;
    }
    else					// nope.. maybe the line
    {						// intersects with left
    	hone = xmin;				// window border?
    	vone = (int) (a * xmin + b);
    	if (vone < ymin || vone > ymax)		// nope.. lower/upper?
    	{
    	    vone = (vone < ymin) ? ymin : ymax;
	    hone = (int) ((vone - b) / a);
	    if (hone < xmin || hone > xmax)
	    	return;
    	}
    }

   // determine right starting point
    if ( (h2 >= xmin && h2 <= xmax) &&		// first estimate:
    	 (v2 >= ymin && v2 <= ymax)		// maybe we can use (h2,v2)
       )					// as left starting point?
    {
    	htwo = h2;
    	vtwo = v2;
    }
    else					// nope.. maybe the line
    {						// intersects with right
    	htwo = xmax;				// window border?
    	vtwo = (int) (a * xmax + b);
    	if (vtwo < ymin || vtwo > ymax)		// nope.. let's see, upper 
    	{					// or lower window border?
	    vtwo = (vtwo < ymin) ? ymin : ymax;
	    htwo = (int) ((vtwo - b) / a);
    	    if (htwo < xmin || htwo > xmax)	// or beyond my reach?
    	    	return;
  	}
    }
    
    // finally...
    Graph::setcolor (foreground_color);
    Graph::drawline (x1 + hone, y1 + vone, x1 + htwo, y1 + vtwo);
}         
